# -*- coding: utf-8 -*-

# Standard Library
import cPickle
from datetime import date
from itertools import chain
from dateutil.relativedelta import relativedelta


# Vendor Library
from django.core.cache import cache
from django.db.models import Q
from django.utils.functional import cached_property

# Project Library
from kinder.core.group.models import Group
from kinder.core.group.models import Pupil

from kinder.core.group.models import GroupStatusEnum
from kinder.core.unit.models import FilialData
from kinder.core.unit.models import FilialTypeStatus
from kinder.core.dict.models import GroupTypeEnumerate
from kinder.core.dict.models import HNE
from kinder.core.direct.models import Direct
from kinder.core.direct.models import DRS
from gisdo.index import BaseIndex
from gisdo.algorithm import (
    EnrolledCountByFamily, EnrolledCountByHealth,
    EnrolledCountShortStay, EnrolledCountFullDayStay,
    EnrolledCountByCorrect, EnrolledCountByCare,
    EnrolledCountWantChangeDou, EnrolledCountByAge,
    EnrolledCountWantChangeDouWithHN, PredictableDecreaseContingent,
    EnrolledCountByDisabled)
from gisdo.constants import GOVERNMENT, DOU_TYPE_MAP, ALL_UNIT_TYPES
from gisdo import settings as gs_settings
from gisdo.utils import AgeDeltas
from .constants import short_day_types


class EnrolledIndex(BaseIndex):
    """
    Запросы группы "Количество детей зачисленных ... "
    """

    COUNT_BY_AGE = 1
    COUNT_DISABLED = 2
    COUNT_FAMILY = 3
    COUNT_CORRECT = 4
    COUNT_HEALTH = 5
    COUNT_SHORT_STAY = 6
    COUNT_FULL_DAY_STAY = 11
    COUNT_WANT_CHANGE_DOU = 7
    COUNT_WANT_CHANGE_DOU_WITH_HN = 8
    COUNT_BY_CARE = 9
    COUNT_DECREASE_CONTINGENT = 10

    ALGORITHM_MAP = {
        COUNT_BY_AGE: EnrolledCountByAge,
        COUNT_DISABLED: EnrolledCountByDisabled,
        COUNT_FAMILY: EnrolledCountByFamily,
        COUNT_CORRECT: EnrolledCountByCorrect,
        COUNT_HEALTH: EnrolledCountByHealth,
        COUNT_SHORT_STAY: EnrolledCountShortStay,
        COUNT_FULL_DAY_STAY: EnrolledCountFullDayStay,
        COUNT_WANT_CHANGE_DOU: EnrolledCountWantChangeDou,
        COUNT_WANT_CHANGE_DOU_WITH_HN: EnrolledCountWantChangeDouWithHN,
        COUNT_BY_CARE: EnrolledCountByCare,
        COUNT_DECREASE_CONTINGENT: PredictableDecreaseContingent
    }

    def get_count(
            self, dou_type=ALL_UNIT_TYPES,
            index_type=COUNT_BY_AGE, **kwargs):

        if dou_type != ALL_UNIT_TYPES:
            if ((not self.dou_type and dou_type != GOVERNMENT) or
                    (self.dou_type and
                     self.dou_type.code not in DOU_TYPE_MAP[dou_type])):
                return 0

        result_set = (
            EnrolledIndex.get_algorithm(
                index_type)(unit=self.unit, session=self.session)
            .get_result_set(**kwargs))

        return self.session.query(result_set)


def get_count(collection, age_filter):
    u"""Определение количества записей в коллекции подходящих под фильтр.

    Функция применяет фильтр age_filter к коллекции collection и возвращает
    количество отфильтрованных записей.
    """

    return len([
        1
        for _, date_of_birth in collection
        if age_filter(date_of_birth)
    ])


def get_age_filter(age, on_date):
    u"""Функция строит фильтр возраста."""

    down, up = AgeDeltas.get_category_deltas(age, on_date)

    filter_ = lambda rec: rec > down
    if up:
        if age == u'от 2 месяцев до 6 месяцев':
            filter_ = lambda rec: (down < rec)
        else:
            filter_ = lambda rec: (down < rec <= up)

    return filter_


def memoize(f):
    u"""Декоратор для запоминания значений функций.

    В декарируемую функцию можно передавать флаг для
    принудительного перерасчета.
    """

    def wrapper(instance, *args, **kwargs):

        if "force" in kwargs:
            force = kwargs["force"]
            del kwargs["force"]
        else:
            force = True

        key = f.__name__, cPickle.dumps(kwargs)

        stored_value = cache.get(key)
        if force or stored_value is None:
            result = f(instance, *args, **kwargs)
            cache.set(key, result)
        else:
            result = stored_value

        return result

    return wrapper


class Index(object):
    u"""Класс для вычисления показателей по группам."""

    def __init__(self, dou):
        self.dou = dou

    @cached_property
    def groups(self):
        u"""Группы данного учреждения.

        Включает группы подтвержденных филиалов.
        """
        branches = FilialData.objects.filter(
            head=self.dou, filial__closed=False,
            status=FilialTypeStatus.APPROVE,
            filial__gisdo__not_on_federal_report=False
        ).values_list("filial_id", flat=True)

        return Group.extended_objects.with_count_norm().filter(
            status=GroupStatusEnum.FACT
        ).filter(
            Q(unit=self.dou) | Q(unit__in=branches)
        ).order_by("id")

    @memoize
    def enrolled_in_group(self, group_id, query_filter):
        u"""Количество зачисленных в группе.
        :param group_id: Идентификатор группы.
        :param query_filter: Название фильтра для выборки
        """

        filter_ = filters.get(query_filter, 'default')()

        group = Group.objects.get(id=group_id)

        children_in_fact = Pupil.objects.filter(
            grup=group, temporary_deduct=False,
            grup__status=GroupStatusEnum.FACT
        ).filter(filter_)

        children_in_plan = Pupil.objects.filter(
            Q(grup__room_id__isnull=False, grup__room_id=group.room_id)
        ).filter(
            temporary_deduct=False, grup__status=GroupStatusEnum.PLAN
        ).filter(
            filter_
        ).exclude(
            children_id__in=Pupil.objects.filter(
                grup__unit=group.unit, temporary_deduct=False,
                grup__status=GroupStatusEnum.FACT
            ).values_list("children_id", flat=True)
        )

        return list(
            chain(
                children_in_fact.values_list(
                    "children_id", "children__date_of_birth"),
                children_in_plan.values_list(
                    "children_id", "children__date_of_birth")
            )
        )

    @memoize
    def directs_count_in_group(self, group_id):
        u"""Количество направлений в группу, влияющих на места."""

        group = Group.objects.get(id=group_id)

        direct_in_group = Direct.objects.filter(
            Q(group=group) &
            Q(status__code__in=[DRS.REGISTER, DRS.DOGOVOR])
        )

        bounded_plan_groups = Group.objects.filter(
            Q(room_id__isnull=False, room_id=group.room_id))
        direct_count_in_bounded_group = Direct.objects.filter(
            Q(group__in=bounded_plan_groups) &
            Q(status__code__in=[DRS.REGISTER, DRS.DOGOVOR])
        ).exclude(
            declaration__children_id__in=Direct.objects.filter(
                Q(group__unit=group.unit) &
                Q(status__code__in=[DRS.REGISTER, DRS.DOGOVOR])
            ).values_list("declaration__children_id", flat=True)
        ).count()

        return direct_in_group.count() + direct_count_in_bounded_group

    @memoize
    def get_capacity_in_group(self, group_id):
        u"""Наполняемость в группе."""

        group = Group.objects.get(id=group_id)

        # определяем тип наполняемости
        if gs_settings.USE_MAX_OCCUPANCY is None:
            is_max_occupancy = group.unit.get_use_fact_norm_cnt()
        else:
            is_max_occupancy = gs_settings.USE_MAX_OCCUPANCY

        if is_max_occupancy:
            return group.max_count or 0
        else:
            try:
                return int(group.count_norm)
            except (TypeError, ValueError):
                return 0

    def free_space_in_group(self, group, force=True):
        u"""Свободные места в группе."""
        occupancy = self.get_capacity_in_group(group.id)
        enrolled = self.enrolled_in_group(
            force=force, group_id=group.id, query_filter='default')
        directs = self.directs_count_in_group(group.id)

        free_space = occupancy - len(list(enrolled)) - directs

        return free_space if free_space > 0 else 0

    def __call__(
            self, force=False, group_filter=None, query_filter="default"):

        if group_filter is None:
            def group_filter(_): return True

        summary = chain()
        for group in self.groups:
            if group_filter(group):
                summary = chain(
                    summary,
                    self.enrolled_in_group(
                        force=force,
                        group_id=group.id,
                        query_filter=query_filter
                    ))

        return summary

    def free_space_in_unit(self, force, group_filter='default'):

        group_filter = free_space_filters.get(group_filter, lambda _: True)

        free_space = {}
        for group in self.groups:
            if not group_filter(group):
                continue
            free_space[group] = self.free_space_in_group(group, force=force)

        return free_space


def predictable_decrease_contingent_filter():
    today = date.today()

    # Если ДР раньше этой даты, то ребенку > 6,5 лет.
    date_of_birth_border = today - relativedelta(years=6, months=6)

    return Q(children__date_of_birth__lt=date_of_birth_border)


filters = {
    'default': lambda: Q(),
    'predictable_decrease_contingent': predictable_decrease_contingent_filter,
    'without_hn': lambda: (
        Q(children__health_need_id__isnull=False) & ~Q(
            children__health_need__code=HNE.NOT
        )
    )
}


free_space_filters = {
    'short_stay': lambda group: True
    if group.work_type and group.work_type.code in short_day_types
    else False,
    'with_health_need': lambda group: True
    if group.type and group.type.code in (GroupTypeEnumerate.COMBI,
                                          GroupTypeEnumerate.COMP)
    else False,
    'compensating': lambda group: True
    if group.type and group.type.code == GroupTypeEnumerate.COMP else False,
    'health_group': lambda group: True
    if group.type and group.type.code == GroupTypeEnumerate.HEALTH else False
}
